# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import relationships
from restalchemy.dm import types


class VM(models.ModelWithUUID):
    state = properties.property(
        types.String(max_length=10), required=True, default="off"
    )
    name = properties.property(
        types.String(max_length=255), required=True, example="testname"
    )
    just_none = properties.property(
        types.AllowNone(types.String()), required=False, default=None
    )
    status = properties.property(
        types.Enum(["active", "disabled"]), default="active", required=True
    )
    created = properties.property(
        types.UTCDateTimeZ(),
        default=lambda: types.DEFAULT_DATE_Z,
    )
    updated = properties.property(
        types.UTCDateTime(),
        default=lambda: types.DEFAULT_DATE,
    )


class Port(models.CustomPropertiesMixin, models.ModelWithUUID):
    __custom_properties__ = {
        "never_call": types.String(),
        "_hidden_field": types.String(),
        "some_field1": types.String(),
        "some_field2": types.String(),
        "some_field3": types.String(),
        "some_field4": types.String(),
        "some_field5": types.AllowNone(types.String()),
        "unique_field": types.String(),
    }

    mac = properties.property(types.Mac(), default="00:00:00:00:00:00")
    vm = relationships.relationship(VM, required=True)

    @property
    def never_call(self):
        raise NotImplementedError("Should be call never")

    @property
    def _hidden_field(self):
        return "_hidden_field"

    @property
    def some_field1(self):
        return "some_field1"

    @property
    def some_field2(self):
        return "some_field2"

    @property
    def some_field3(self):
        return "some_field3"

    @property
    def some_field4(self):
        return "some_field4"

    @property
    def some_field5(self):
        return None

    @property
    def unique_field(self):
        return str(self.uuid)


class IpAddress(models.ModelWithUUID):
    ip = properties.property(types.String(), default="192.168.0.1")
    port = relationships.relationship(Port, required=True)


class Tag(models.ModelWithUUID):
    name = properties.property(types.String(), id_property=True, required=True)
    visible = properties.property(types.Boolean(), default=True, required=True)

    @classmethod
    def get_id_property(cls):
        return {"name": cls.id_properties["name"]}



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg

from restalchemy.common import config
from restalchemy.common import config_opts
from restalchemy.common import log as ra_log
from restalchemy.storage.sql import engines
from restalchemy.storage.sql import migrations

cmd_opts = [
    cfg.StrOpt(
        "migration",
        short="m",
        required=True,
        help="migrate to",
    ),
    cfg.StrOpt(
        "path",
        required=True,
        short="p",
        help="Path to migrations folder",
    ),
    cfg.BoolOpt(
        "dry-run",
        default=False,
        help="Dry run downgrade migration w/o any real changes.",
    ),
]

CONF = cfg.CONF
CONF.register_cli_opts(cmd_opts)
config_opts.register_common_db_opts(conf=CONF)


def main():
    config.parse(sys.argv[1:])
    ra_log.configure()
    engines.engine_factory.configure_factory(db_url=CONF.db.connection_url)
    engine = migrations.MigrationEngine(migrations_path=CONF.path)
    engine.rollback_migration(migration_name=CONF.migration, dry_run=CONF.dry_run)



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import contextlib
import logging
import urllib.parse as parse

from mysql.connector import pooling
import psycopg_pool
from psycopg.types.json import Jsonb, JsonbDumper

from restalchemy.common import constants as c
from restalchemy.common import singletons
from restalchemy.storage.sql.dialect import adapters
from restalchemy.storage.sql.dialect import mysql
from restalchemy.storage.sql.dialect import pgsql
from restalchemy.storage.sql import sessions

DEFAULT_NAME = "default"
DEFAULT_CONNECTION_TIMEOUT = 10
LOG = logging.getLogger(__name__)


class AbstractEngine(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def URL_SCHEMA(self):
        """
        The URL schema for the database engine.

        This abstract property should be overridden by subclasses to specify
        the schema used in the database connection URL (e.g., "postgresql").
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def DEFAULT_PORT(self):
        """The default port number used by the database engine.

        This abstract property should be overridden by subclasses to specify
        the default port number used by the database engine if it is not
        specified in the database connection URL.
        """
        raise NotImplementedError()

    def __init__(
        self,
        db_url,
        dialect,
        session_storage,
        config=None,
        query_cache=False,
    ):
        """
        Initializes the database engine.

        :param db_url: The URL of the database connection.
        :param dialect: The SQL dialect used by the database engine.
        :param session_storage: The type of session storage used by the engine.
        :param config: A dictionary of configuration options for the engine.
        :param query_cache: A boolean indicating whether the engine should
                            cache query results.

        :raises ValueError: If the database URL does not match the expected
            format.
        """
        super(AbstractEngine, self).__init__()
        self._db_url = parse.urlparse(db_url)

        if self._db_url.scheme != self.URL_SCHEMA:
            raise ValueError(
                "Database url should be starts with "
                f"{self.URL_SCHEMA}://. For example: "
                "mysql://<username>:[password]@"
                "<host>:[port]/<database_name>"
            )

        self._db_name = self._db_url.path[1:]
        self._config = config or {}
        self._dialect = dialect
        self._session_storage = session_storage
        self._query_cache = query_cache

    @property
    def dialect(self):
        """
        Returns the SQL dialect used by the database engine.

        :return: The SQL dialect used by the database engine.
        :rtype: AbstractDialect
        """
        return self._dialect

    def escape(self, value):
        """
        Escapes a value for use in a query.

        Escapes a value by wrapping it in backticks. This is a common way to
        escape values in SQL queries.

        :param value: The value to be escaped.
        :return: The escaped value.
        :rtype: str
        """
        return "`%s`" % value

    @property
    def db_name(self):
        """
        Returns the name of the database.

        This property retrieves the database name from the parsed URL of the
        database connection, which is stored during the initialization of the
        engine instance.
        """

        return self._db_name

    @property
    def db_username(self):
        """
        Returns the username used for the database connection.

        This property retrieves the username from the parsed URL of the
        database connection, which is stored during the initialization of the
        engine instance.
        """
        return self._db_url.username

    @property
    def db_password(self):
        """
        Returns the password used for the database connection.

        This property retrieves the password from the parsed URL of the
        database connection, which is stored during the initialization of the
        engine instance.
        """
        return self._db_url.password

    @property
    def db_host(self):
        """
        Returns the hostname used for the database connection.

        This property retrieves the hostname from the parsed URL of the
        database connection, which is stored during the initialization of the
        engine instance.
        """
        return self._db_url.hostname

    @property
    def db_port(self):
        """
        Returns the port number used for the database connection.

        If the port number was not specified in the database connection URL,
        the default port number for the engine is used instead.

        Returns:
            The port number used for the database connection.
        """
        return self._db_url.port or self.DEFAULT_PORT

    @property
    def query_cache(self):
        """
        Returns a boolean indicating whether the query cache is enabled.

        The query cache can be set when the engine is initialized, and it
        determines whether the engine should cache the results of queries
        using the session's query cache.

        Returns:
            A boolean indicating whether the query cache is enabled.
        """
        return self._query_cache

    def get_connection(self):
        """
        Establishes and returns a connection to the database from a pool.

        This method should be implemented by subclasses to provide the logic
        for connecting to a specific database using the engine's
        configuration.

        Returns:
            A database connection object.

        Raises:
            NotImplementedError: If the method is not implemented by a
            subclass.
        """
        raise NotImplementedError()

    def get_session(self):
        """
        Returns a session object for the engine. This method can be used to
        explicitly start a session, or to get the session object from the
        current context.

        Returns:
            A session object.

        Raises:
            NotImplementedError: If the method is not implemented by a
            subclass.
        """
        raise NotImplementedError()

    def _get_session_from_storage(self):
        """
        Retrieve a session object from the session storage associated with
        this engine.

        :returns: A session object, or None if no session is stored.
        :rtype: Session or None
        """
        try:
            return self.get_session_storage().get_session()
        except sessions.SessionNotFound:
            return None

    @contextlib.contextmanager
    def session_manager(self, session=None):
        """
        Context manager for managing a database session.

        This method provides a context for executing database operations within
        a session. If a session is not provided, it attempts to retrieve one
        from storage or creates a new session. The session is automatically
        committed if no exceptions occur, otherwise it is rolled back. The
        session is always closed when the context exits.

        :param session: An optional session object. If not provided, a session
            is retrieved from storage or a new one is created.
        :type session: Optional[Session]
        :yields: The active session for database operations.
        :raises Exception: Any exceptions raised during the session's execution
            will result in a rollback of the session.
        """

        session = session or self._get_session_from_storage()
        if session is None:
            session = self.get_session()
            try:
                yield session
                session.commit()
            except Exception:
                session.rollback()
                raise
            finally:
                session.close()
        else:
            yield session

    def get_session_storage(self):
        """
        Returns the session storage object for the engine.

        The session storage object is used to store session objects for the
        engine. This allows the engine to maintain a consistent session
        object across different threads and contexts.

        Returns:
            The session storage object.
        """
        return self._session_storage

    def close_connection(self, conn):
        """
        Closes the provided connection.

        This method closes a database connection. It is typically used to
        release resources associated with the connection once database
        operations are completed.

        :param conn: The connection to be closed.
        :type conn: Connection object
        """

        conn.close()


class PgDictJsonbDumper(JsonbDumper):
    def dump(self, obj):
        return obj if obj is None else super().dump(Jsonb(obj))


class PgSQLEngine(AbstractEngine):
    URL_SCHEMA = c.RA_POSTGRESQL_PROTO_NAME
    DEFAULT_PORT = c.RA_POSTGRESQL_DB_PORT

    def __init__(self, db_url, config=None, query_cache=False):
        """
        Initializes the PostgreSQL engine.

        :param db_url: The connection URL for the PostgreSQL database.
        :param config: A dictionary of configuration options for the engine.
        :param query_cache: A boolean indicating whether the engine should
            cache query results.

        :return: The initialized engine.
        """

        super(PgSQLEngine, self).__init__(
            db_url=db_url,
            dialect=pgsql.PgSQLDialect(),
            session_storage=sessions.SessionThreadStorage(),
            config=config,
            query_cache=query_cache,
        )

        # RA expects the pool to be ready to use
        if "open" not in self._config:
            self._config["open"] = True

        self._pool = psycopg_pool.ConnectionPool(
            conninfo=db_url,
            configure=self._conn_configure_callback,
            **self._config,
        )
        self._pool.wait()

    def _conn_configure_callback(self, conn):
        conn.adapters.register_dumper(dict, PgDictJsonbDumper)

    def escape(self, value):
        """
        Escapes a value for use in a PostgreSQL query.

        This method is used by the :py:class:`PgSQLSession` class to escape
        values before they are used in a query. It is not intended to be used
        directly by applications.

        The value is escaped by wrapping it in double quotes. This is because
        double quotes are used to delimit strings in PostgreSQL, and the value
        is already escaped if it is a string.

        :param value: The value to be escaped.
        :return: The escaped value.
        """

        return '"' + value + '"'

    def __del__(self):
        """
        Closes the connection pool to the PostgreSQL database before the
        engine instance is garbage collected.

        This method is called when the engine instance is garbage collected
        to ensure that the connection pool is closed and any resources
        acquired by the pool are released.

        Note that the connection pool is not closed until the engine instance
        is garbage collected, which may not happen immediately. Therefore,
        this method is not suitable for use in applications where it is
        important to release resources immediately after they are no longer
        needed.

        :raises psycopg2.Error: If an error occurs while closing the pool.
        """

        self._pool.close()

    def get_session(self):
        """
        Creates and returns a new PgSQLSession instance for the current engine.

        This method initializes a session object that can be used to interact
        with the PostgreSQL database associated with this engine.

        :returns: A session object for database operations.
        :rtype: PgSQLSession
        """

        return sessions.PgSQLSession(engine=self)

    def get_connection(self):
        """
        Retrieves a connection from the pool.

        This method is used to acquire a PostgreSQL connection from the pool
        of connections managed by the engine. The connection is used to
        execute queries and other database operations.

        :returns: A connection object for database operations.
        :rtype: psycopg2.extensions.connection
        """

        return self._pool.getconn()

    def close_connection(self, conn):
        """
        Releases a connection back to the pool.

        This method runs appropriate cleanup routines and returns
        the connection back to the pool for reuse.

        :param conn: The connection to be released back to the pool.
        :type conn: psycopg2.extensions.connection
        """

        self._pool.putconn(conn)


class MySQLEngine(AbstractEngine):
    URL_SCHEMA = c.RA_MYSQL_PROTO_NAME
    DEFAULT_PORT = c.RA_MYSQL_DB_PORT

    def __init__(self, db_url, config=None, query_cache=False):
        """
        Initializes the MySQL engine.

        :param db_url: The URL of the database connection.
        :param config: A dictionary of configuration options for the engine.
        :param query_cache: A boolean indicating whether the engine should
            cache query results.

        :raises ValueError: If the database URL does not match the expected
            format.
        """
        super(MySQLEngine, self).__init__(
            db_url=db_url,
            dialect=mysql.MySQLDialect(),
            session_storage=sessions.SessionThreadStorage(),
            config=config,
            query_cache=query_cache,
        )

        if "connection_timeout" not in self._config:
            self._config["connection_timeout"] = DEFAULT_CONNECTION_TIMEOUT
        self._config.update(
            {
                "user": self.db_username,
                "password": self.db_password,
                "database": self.db_name,
                "host": self.db_host,
                "port": self.db_port,
                "converter_class": adapters.MySQLConverter,
            }
        )

        try:
            self._pool = pooling.MySQLConnectionPool(**self._config)
        except AttributeError as e:
            pool_name = e.args[0].split("'")[1]
            new_name = str(hash(pool_name))
            LOG.warning("Changing '%s' pool name to '%s'", pool_name, new_name)
            config["pool_name"] = new_name
            self._pool = pooling.MySQLConnectionPool(**self._config)

    def __del__(self):
        """
        Closes the connection pool to the MySQL database before the engine
        instance is garbage collected.

        This method is called when the engine instance is garbage collected to
        ensure that the connection pool is closed and any resources acquired by
        the pool are released.
        """
        pool = getattr(self, "_pool", None)
        if pool is not None:
            self._pool._remove_connections()

    def get_connection(self):
        """
        Retrieves a connection from the pool.

        This method is used to acquire a connection from the pool of
        connections managed by the engine. The connection is used to
        execute queries and other database operations.

        :returns: A connection object for database operations.
        :rtype: mysql.connector.connection.MySQLConnection
        """
        return self._pool.get_connection()

    def get_session(self):
        """
        Creates and returns a new MySQLSession instance for the current engine.

        This method initializes a session object that can be used to interact
        with the MySQL database associated with this engine.

        :returns: A session object for database operations.
        :rtype: MySQLSession
        """
        return sessions.MySQLSession(engine=self)


class EngineFactory(singletons.InheritSingleton):
    def __init__(self):
        """
        Initializes the engine factory singleton.

        This method is called when the singleton is created. It initializes the
        internal state of the factory by creating an empty dictionary to store
        engine instances and a mapping of database URLs to engine classes.

        :returns: None
        """
        super(EngineFactory, self).__init__()
        self._engines = {}
        self._engines_map = {
            MySQLEngine.URL_SCHEMA: MySQLEngine,
            PgSQLEngine.URL_SCHEMA: PgSQLEngine,
        }

    def configure_postgresql_factory(
        self,
        conf,
        section=c.DB_CONFIG_SECTION,
        name=DEFAULT_NAME,
    ):
        """
        Configures the engine factory for a PostgreSQL database.

        This method is a convenience wrapper around the `configure_factory`
        method that is specific to PostgreSQL databases. It takes the same
        parameters as `configure_factory` but provides default values for the
        configuration options that are specific to PostgreSQL.

        :param conf: The configuration object to read from.
        :param section: The section of the configuration object to read from.
        :param name: The name of the engine to configure.
        """
        self.configure_factory(
            db_url=conf[section].connection_url,
            config={
                "min_size": conf[section].connection_pool_min_size,
                "max_size": conf[section].connection_pool_max_size,
                "open": conf[section].connection_pool_open,
                "timeout": conf[section].connection_pool_client_timeout,
                "max_waiting": conf[section].connection_pool_max_waiting,
                "max_lifetime": conf[section].connection_max_lifetime,
                "max_idle": conf[section].connection_max_idle,
                "reconnect_timeout": conf[section].connection_pool_reconnect_timeout,
                "num_workers": conf[section].connection_pool_num_workers,
            },
            query_cache=conf[section].connection_query_cache,
            name=name,
        )

    def configure_mysql_factory(
        self,
        conf,
        section=c.DB_CONFIG_SECTION,
        name=DEFAULT_NAME,
    ):
        """
        Configures the engine factory for a MySQL database.

        This method is a convenience wrapper around the `configure_factory`
        method that is specific to MySQL databases. It takes the same
        parameters as `configure_factory` but provides default values for the
        configuration options that are specific to MySQL.

        :param conf: The configuration object to read from.
        :param section: The section of the configuration object to read from.
        :param name: The name of the engine to configure.
        """
        self.configure_factory(
            db_url=conf[section].connection_url,
            config={
                "pool_size": conf[section].connection_pool_size,
            },
            query_cache=conf[section].connection_query_cache,
            name=name,
        )

    def configure_factory(
        self,
        db_url,
        config=None,
        query_cache=False,
        name=DEFAULT_NAME,
    ):
        """
        Configures and creates a new database engine instance for the given
        URL.

        This method initializes a database engine based on the provided
        database URL, configuration options, and query caching preference. The
        initialized engine is stored in the factory's internal engines
        dictionary, keyed by the specified name.

        :param db_url: The database connection URL.
        :param config: A dictionary of configuration options for the engine.
                    Defaults to None.
        :param query_cache: A boolean indicating whether the engine should
                            cache query results. Defaults to False.
        :param name: The name under which to store the initialized engine in
                     the factory. Defaults to 'default'.

        :raises ValueError: If the schema from the db_url is not supported
                            or if no driver is found for the schema.
        """

        schema = db_url.split(":")[0]
        try:
            self._engines[name] = self._engines_map[schema.lower()](
                db_url=db_url, config=config, query_cache=query_cache
            )
        except KeyError:
            raise ValueError("Can not find driver for schema %s" % schema)

    def get_engine(self, name=DEFAULT_NAME):
        """
        Returns an engine instance from the factory's internal engines
        dictionary by name. If the specified engine name does not exist,
        a ValueError is raised.

        :param name: The name of the engine to return. Defaults to 'default'.
        :return: An engine instance.
        :raises ValueError: If the specified engine name is not found in the
                            factory's internal engines dictionary.
        """
        engine = self._engines.get(name, None)
        if engine:
            return engine
        raise ValueError(
            ("Can not return %s engine. Please configure EngineFactory") % name,
        )

    def destroy_engine(self, name=DEFAULT_NAME):
        """
        Removes and destroys the engine instance associated with the specified
        name from the factory's internal engines dictionary.

        This method attempts to delete the engine instance identified by the
        given name. If the engine instance does not exist, the method silently
        handles the KeyError exception.

        :param name: The name of the engine to destroy. Defaults to 'default'.
        """

        try:
            del self._engines[name]
        except KeyError:
            pass

    def destroy_all_engines(self):
        """
        Removes and destroys all engine instances from the factory's internal
        engines dictionary.

        This method resets the factory to its initial state, removing all
        configured engines. Subsequent calls to get_engine() will raise a
        ValueError until at least one engine is configured using
        configure_factory().
        """
        self._engines = {}


class DBConnectionUrl(object):
    _CENSORED = ":<censored>@"

    def __init__(self, db_url):
        """
        Initializes a DBConnectionUrl instance with the provided database URL.

        :param db_url: The URL of the database connection, which will be
                       parsed and stored.
        """

        super(DBConnectionUrl, self).__init__()
        self._db_url = parse.urlparse(db_url)

    def __repr__(self):
        """
        Returns a string representation of the DBConnectionUrl instance.

        The string representation is generated by replacing the password
        substring in the original URL with the value of _CENSORED, to prevent
        the password from being displayed.

        :return: A string representation of the DBConnectionUrl instance.
        """
        if self._db_url.password is None:
            orig_substr = "@"
        else:
            orig_substr = ":%s@" % self._db_url.password
        return self.url.replace(orig_substr, self._CENSORED)

    @property
    def url(self):
        """
        Returns the full URL string of the database connection.

        This property retrieves the complete URL by converting the parsed
        URL components back into a string representation. It is useful for
        obtaining the original connection URL, with the password censored
        if applicable, after it has been parsed and stored.
        """

        return self._db_url.geturl()


engine_factory = EngineFactory()



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from restalchemy.storage.sql import migrations
from restalchemy.tests.functional.storage.mysql.prefetch import models


class MigrationStep(migrations.AbstractMigrationStep):
    def __init__(self):
        self._depends = ["prefetch-relationship-tests-f3841e.py"]

    @property
    def migration_id(self):
        return "9727f325-aab6-4393-8edb-3c66c18ed772"

    def upgrade(self, session):
        root = models.Root()
        root.insert()

        lnp1_1 = models.LNP1_1(root=root)
        lnp1_1.insert()

        lnp1_2 = models.LNP1_2()
        lnp1_2.insert()

        lwp1_1 = models.LWP1_1(root=root)
        lwp1_1.insert()
        lwp1_2 = models.LWP1_2()
        lwp1_2.insert()

        lwp2_1 = models.LWP2_1(lwp1_1=lwp1_1)
        lwp2_1.insert()
        lwp2_2 = models.LWP2_2()
        lwp2_2.insert()
        lwp2_3 = models.LWP2_3(lwp1_2=lwp1_2)
        lwp2_3.insert()
        lwp2_4 = models.LWP2_4()
        lwp2_4.insert()

    def downgrade(self, session):
        for Model in [
            models.LWP2_1,
            models.LWP2_2,
            models.LWP2_3,
            models.LWP2_4,
            models.LWP1_1,
            models.LWP1_2,
            models.LNP1_1,
            models.LNP1_2,
            models.Root,
        ]:
            for model in Model.objects.get_all():
                model.delete()


migration_step = MigrationStep()



#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import multidict

from restalchemy.common import exceptions


class CanNotGetActiveMethod(exceptions.RestAlchemyException):
    message = "Can not get active RA method from API context"


class RequestContext(object):
    _special_params = frozenset(
        ("fields", "page_limit", "page_marker", "sort_key", "sort_dir")
    )

    def __init__(self, request):
        super(RequestContext, self).__init__()
        self._req = request
        self._fields_to_show = request.params.getall("fields")
        self._method = None

    def set_active_method(self, method):
        self._method = method

    def get_active_method(self):
        if self._method:
            return self._method
        raise CanNotGetActiveMethod()

    @property
    def params(self):
        return multidict.MultiDict(self._req.params.items())

    @property
    def params_filters(self):
        result_multi_dict_items = [
            (name, value)
            for name, value in self._req.params.items()
            if name not in self._special_params
        ]
        return multidict.MultiDict(result_multi_dict_items)

    def can_be_shown_field(self, resource_field_name):
        if self._fields_to_show:
            return resource_field_name in self._fields_to_show
        return True



# Copyright 2025
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from restalchemy.storage.sql import utils
from restalchemy.tests.unit import base


class SavepointCtxTestCase(base.BaseTestCase):
    def _make_ctx(self, dialect_name):
        engine = mock.Mock()
        engine.dialect.name = dialect_name

        ctx = mock.Mock()
        ctx._engine = engine
        ctx.get_session.return_value = mock.Mock()
        return ctx

    @mock.patch("restalchemy.storage.sql.utils.contexts.Context")
    def test_savepoint_postgresql_success(self, mock_ctx_cls):
        ctx = self._make_ctx("postgresql")
        mock_ctx_cls.return_value = ctx

        session = ctx.get_session.return_value

        with utils.savepoint("sp") as s:
            self.assertIs(s, session)

        session.execute.assert_any_call("SAVEPOINT sp;", tuple())
        session.execute.assert_any_call("RELEASE SAVEPOINT sp;", tuple())

    @mock.patch("restalchemy.storage.sql.utils.contexts.Context")
    def test_savepoint_postgresql_error(self, mock_ctx_cls):
        ctx = self._make_ctx("postgresql")
        mock_ctx_cls.return_value = ctx

        session = ctx.get_session.return_value

        with self.assertRaises(RuntimeError):
            with utils.savepoint("sp"):
                raise RuntimeError("boom")

        session.execute.assert_any_call("SAVEPOINT sp;", tuple())
        session.execute.assert_any_call("ROLLBACK TO SAVEPOINT sp;", tuple())
        session.execute.assert_any_call("RELEASE SAVEPOINT sp;", tuple())

    @mock.patch("restalchemy.storage.sql.utils.contexts.Context")
    def test_savepoint_mysql_success(self, mock_ctx_cls):
        ctx = self._make_ctx("mysql")
        mock_ctx_cls.return_value = ctx

        session = ctx.get_session.return_value

        with utils.savepoint("sp") as s:
            self.assertIs(s, session)

        session.execute.assert_any_call("SAVEPOINT sp;", tuple())
        session.execute.assert_any_call("RELEASE SAVEPOINT sp;", tuple())

    @mock.patch("restalchemy.storage.sql.utils.contexts.Context")
    def test_savepoint_mysql_error(self, mock_ctx_cls):
        ctx = self._make_ctx("mysql")
        mock_ctx_cls.return_value = ctx

        session = ctx.get_session.return_value

        with self.assertRaises(RuntimeError):
            with utils.savepoint("sp"):
                raise RuntimeError("boom")

        session.execute.assert_any_call("SAVEPOINT sp;", tuple())
        session.execute.assert_any_call("ROLLBACK TO sp;", tuple())
        session.execute.assert_any_call("RELEASE SAVEPOINT sp;", tuple())



#    Copyright 2021 Eugene Frolov <eugene@frolov.net.ru>.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.tests.functional import base
from restalchemy.tests.functional.storage.mysql.prefetch import models


class PrefetchTestCase(base.BaseWithDbMigrationsTestCase):
    __LAST_MIGRATION__ = "prefetch-relationship-tests-data-9727f3"
    __FIRST_MIGRATION__ = "prefetch-relationship-tests-f3841e"

    def setUp(self):
        super(PrefetchTestCase, self).setUp()

        models.OBJECT_COLLECTION_MOCK.reset_mock()

    def test_get_model_without_relationships(self):

        results = models.Root.objects.get_all()

        self.assertEqual(1, len(results))
        self.assertEqual(1, models.OBJECT_COLLECTION_MOCK.get_all.call_count)

    def test_get_model_with_relationships_and_prefetch_false(self):
        results = models.LNP1_1.objects.get_all()

        self.assertEqual(1, len(results))
        self.assertEqual(2, models.OBJECT_COLLECTION_MOCK.get_all.call_count)

    def test_get_model_with_relationships_and_prefetch_true(self):
        results = models.LWP1_1.objects.get_all()

        self.assertEqual(1, len(results))
        self.assertEqual(1, models.OBJECT_COLLECTION_MOCK.get_all.call_count)

    def test_get_model_with_deep_relationships_and_prefetch_true(self):
        results = models.LWP2_1.objects.get_all()

        self.assertEqual(1, len(results))
        self.assertIsInstance(results[0], models.LWP2_1)
        self.assertIsInstance(results[0].lwp1_1, models.LWP1_1)
        self.assertIsInstance(results[0].lwp1_1.root, models.Root)
        self.assertEqual(1, models.OBJECT_COLLECTION_MOCK.get_all.call_count)

    def test_get_model_with_relationships_is_none_and_prefetch_true(self):
        results = models.LWP1_2.objects.get_all()

        self.assertEqual(1, len(results))
        self.assertIsInstance(results[0], models.LWP1_2)
        self.assertIsInstance(results[0].root, type(None))
        self.assertEqual(1, models.OBJECT_COLLECTION_MOCK.get_all.call_count)



# Copyright (c) 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc

from restalchemy.common import exceptions as exc
from restalchemy.dm import models
from restalchemy.dm import properties


def relationship(property_type, *args, **kwargs):
    prefetch = kwargs.get("prefetch", False)
    for arg in args:
        if not issubclass(arg, models.Model):
            raise exc.RelationshipModelError(model=arg)
    kwargs["property_class"] = kwargs.get(
        "property_class",
        PrefetchRelationship if prefetch else Relationship,
    )
    return properties.property(property_type=property_type, *args, **kwargs)


def required_relationship(property_type, *args, **kwargs):
    kwargs["required"] = True
    return relationship(property_type, *args, **kwargs)


def readonly_relationship(property_type, *args, **kwargs):
    kwargs["read_only"] = True
    return required_relationship(property_type, *args, **kwargs)


class BaseRelationship(properties.AbstractProperty, metaclass=abc.ABCMeta):
    pass


class Relationship(BaseRelationship):
    def __init__(
        self,
        property_type,
        default=None,
        required=False,
        read_only=False,
        value=None,
    ):
        if value and not isinstance(value, property_type):
            raise TypeError("Expected '%s' type; value: %r" % (property_type, value))
        self._type = property_type
        self._required = bool(required)
        self._read_only = bool(read_only)
        self._default = None
        if default:
            default = default() if callable(default) else default
            self._default = (
                (default()) if callable(default) else self._safe_value(default)
            )
        self._value = value
        self.__first_value = self.value

    def is_dirty(self):
        return not self.__first_value == self.value

    def _safe_value(self, value):
        if value is None or isinstance(value, self._type):
            if value is None and self.is_required():
                raise exc.PropertyRequired()
            return value
        else:
            raise exc.TypeError(value=value, property_type=self._type)

    def is_read_only(self):
        return self._read_only

    def is_required(self):
        return self._required

    @classmethod
    def is_prefetch(cls):
        return False

    @property
    def value(self):
        return self._value or self._default

    @value.setter
    def value(self, value):
        if self.is_read_only():
            raise exc.ReadOnlyProperty()
        self._value = self._safe_value(value)

    def set_value_force(self, value):
        self._value = self._safe_value(value)

    @property
    def property_type(self):
        return self._type

    def get_property_type(self):
        return self._type

    @classmethod
    def is_id_property(cls):
        return False


class PrefetchRelationship(Relationship):
    @classmethod
    def is_prefetch(cls):
        return True



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg

from restalchemy.common import constants as c

CONF = cfg.CONF


def register_common_db_opts(
    conf,
    connection_url=None,
    config_section=c.DB_CONFIG_SECTION,
):
    """
    Registers common database options

    :param conf: Config instance
    :param connection_url: default connection url
    :param config_section: section name to register options in
    """
    db_opt = [
        cfg.StrOpt(
            name="connection_url",
            default=connection_url,
            help="Connection URL to db",
        ),
        cfg.BoolOpt(
            "connection_query_cache",
            default=True,
            help="Cache queries to the database within the transaction",
        ),
        cfg.StrOpt(
            "migrations_path",
            default="migrations",
            help="Path to db migrations",
        ),
        cfg.BoolOpt(
            "apply_migrations_on_startup",
            default=True,
            help="Apply migrations on startup",
        ),
    ]

    conf.register_cli_opts(db_opt, group=config_section)


def register_posgresql_db_opts(
    conf=CONF,
    username=c.RA_POSTGRESQL_USERNAME,
    password=c.RA_POSTGRESQL_PASSWORD,
    host=c.RA_POSTGRESQL_DB_HOST,
    port=c.RA_POSTGRESQL_DB_PORT,
    db_name=c.RA_POSTGRESQL_DB_NAME,
    config_section=c.DB_CONFIG_SECTION,
):
    """
    Registers the configuration options required for a PostgreSQL DB connection.

    This includes the common database options and specific options for the
    PostgreSQL connection pool such as minimum and maximum pool size, connection
    timeout, and other pool management options.

    :param conf: The configuration object to register the options with.
    :param username: The username to use for the connection.
    :param password: The password to use for the connection.
    :param host: The host to connect to.
    :param port: The port to connect to.
    :param db_name: The name of the database to connect to.
    :param config_section: The configuration section to register the options in.
    """

    proto = c.RA_POSTGRESQL_PROTO_NAME
    connection_url = f"{proto}://{username}:{password}@{host}:{port}/{db_name}"

    register_common_db_opts(
        conf=conf,
        connection_url=connection_url,
        config_section=config_section,
    )

    db_opt = [
        cfg.IntOpt(
            name="connection_pool_min_size",
            default=1,
            help=(
                "The minimum number of connection the pool will hold. The"
                " pool will actively try to create new connections if some"
                " are lost (closed, broken) and will try to never go below"
                " min_size"
            ),
        ),
        cfg.IntOpt(
            name="connection_pool_max_size",
            default=2,
            help=(
                "The maximum number of connections the pool will hold. If"
                " None, or equal to min_size, the pool will not grow or"
                " shrink. If larger than min_size, the pool can grow if more"
                " than min_size connections are requested at the same time"
                " and will shrink back after the extra connections have been"
                " unused for more than max_idle seconds."
            ),
        ),
        cfg.BoolOpt(
            "connection_pool_open",
            default=True,
            help=(
                " If True, open the pool, creating the required connections,"
                " on init. If False, open the pool when open() is called"
                " or when the pool context is entered. See the open() method"
                " documentation for more details."
            ),
        ),
        cfg.FloatOpt(
            "connection_pool_client_timeout",
            default=30.0,
            help=(
                "The default maximum time in seconds that a client can wait"
                " to receive a connection from the pool (using connection()"
                " or getconn()). Note that these methods allow to override"
                " the timeout default."
            ),
        ),
        cfg.IntOpt(
            "connection_pool_max_waiting",
            default=10,
            help=(
                "Maximum number of requests that can be queued to the pool,"
                " after which new requests will fail, raising"
                " TooManyRequests. 0 means no queue limit."
            ),
        ),
        cfg.FloatOpt(
            "connection_max_lifetime",
            default=3600.0,
            help=(
                "The maximum lifetime of a connection in the pool, in"
                " seconds. Connections used for longer get closed and"
                " replaced by a new one. The amount is reduced by a"
                " random 10% to avoid mass eviction."
            ),
        ),
        cfg.FloatOpt(
            "connection_max_idle",
            default=600.0,
            help=(
                "Maximum time, in seconds, that a connection can stay"
                " unused in the pool before being closed, and the pool"
                " shrunk. This only happens to connections more than"
                " min_size, if max_size allowed the pool to grow."
            ),
        ),
        cfg.FloatOpt(
            "connection_pool_reconnect_timeout",
            default=300.0,
            help=(
                "Maximum time, in seconds, the pool will try to create"
                " a connection. If a connection attempt fails, the pool"
                " will try to reconnect a few times, using an exponential"
                " backoff and some random factor to avoid mass attempts."
                " If repeated attempts fail, after reconnect_timeout second"
                " the connection attempt is aborted and the"
                " reconnect_failed() callback invoked."
            ),
        ),
        cfg.IntOpt(
            "connection_pool_num_workers",
            default=1,
            help=(
                "Number of background worker threads used to maintain the"
                " pool state. Background workers are used for example to"
                " create new connections and to clean up connections when"
                " they are returned to the pool."
            ),
        ),
    ]

    conf.register_cli_opts(db_opt, config_section)


def register_mysql_db_opts(
    conf=CONF,
    username=c.RA_MYSQL_USERNAME,
    password=c.RA_MYSQL_PASSWORD,
    host=c.RA_MYSQL_DB_HOST,
    port=c.RA_MYSQL_DB_PORT,
    db_name=c.RA_MYSQL_DB_NAME,
    config_section=c.DB_CONFIG_SECTION,
):
    """
    Registers the configuration options required for a MySQL DB connection.

    These are as follows:

    - connection_pool_size: The number of connection the pool will hold.

    :param conf: The configuration object to register the options with.
    :param username: The username to use for the connection.
    :param password: The password to use for the connection.
    :param host: The host to connect to.
    :param port: The port to connect to.
    :param db_name: The name of the database to connect to.
    :param config_section: The configuration section to register the options in.
    """
    proto = c.RA_MYSQL_PROTO_NAME
    connection_url = f"{proto}://{username}:{password}@{host}:{port}/{db_name}"

    register_common_db_opts(
        conf=conf,
        connection_url=connection_url,
        config_section=config_section,
    )

    db_opt = [
        cfg.IntOpt(
            name="connection_pool_size",
            default=1,
            help=(
                "The number of connection the pool will hold. The"
                " pool will actively try to create new connections if some"
                " are lost (closed, broken) and will try to never go below"
                " size"
            ),
        ),
    ]

    conf.register_cli_opts(db_opt, config_section)



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import orjson

from restalchemy.common import exceptions as common_exc
from restalchemy.dm import filters as dm_filters
from restalchemy.dm import models
from restalchemy.storage import base
from restalchemy.storage import exceptions
from restalchemy.storage.sql.dialect import exceptions as exc
from restalchemy.storage.sql import engines
from restalchemy.storage.sql import tables


class ObjectCollection(
    base.AbstractObjectCollection, base.AbstractObjectCollectionCountMixin
):
    @property
    def _table(self):
        return self.model_cls.get_table()

    @property
    def _engine(self):
        return engines.engine_factory.get_engine()

    @base.error_catcher
    def get_all(
        self,
        filters=None,
        session=None,
        cache=False,
        limit=None,
        order_by=None,
        locked=False,
    ):
        with self._engine.session_manager(session=session) as s:
            if cache is True:
                return s.cache.get_all(
                    engine=self._engine,
                    table=self._table,
                    filters=filters,
                    fallback=self._get_all,
                    limit=limit,
                    order_by=order_by,
                    locked=locked,
                )

            return self._get_all(
                filters=filters,
                session=s,
                limit=limit,
                order_by=order_by,
                locked=locked,
            )

    def _get_all(self, filters, session, limit, order_by=None, locked=False):
        result = self._table.select(
            engine=self._engine,
            filters=filters,
            limit=limit,
            order_by=order_by,
            session=session,
            locked=locked,
        )
        return [self.model_cls.restore_from_storage(**params) for params in result.rows]

    @base.error_catcher
    def get_one(self, filters=None, session=None, cache=False, locked=False):
        result = self.get_all(
            filters=filters,
            session=session,
            cache=cache,
            limit=2,
            locked=locked,
        )
        result_len = len(result)
        if result_len == 1:
            return result[0]
        elif result_len == 0:
            raise exceptions.RecordNotFound(model=self.model_cls, filters=filters)
        else:
            raise exceptions.HasManyRecords(model=self.model_cls, filters=filters)

    def get_one_or_none(self, filters=None, session=None, cache=False, locked=False):
        try:
            return self.get_one(
                filters=filters, session=session, cache=cache, locked=locked
            )
        except exceptions.RecordNotFound:
            return None

    def _query(self, where_conditions, where_values, session, limit, order_by, locked):
        result = self._table.custom_select(
            engine=self._engine,
            where_conditions=where_conditions,
            where_values=where_values,
            session=session,
            limit=limit,
            order_by=order_by,
            locked=locked,
        )
        return [
            self.model_cls.restore_from_storage(**params)
            for params in list(result.fetchall())
        ]

    @base.error_catcher
    def query(
        self,
        where_conditions,
        where_values,
        session=None,
        cache=False,
        limit=None,
        order_by=None,
        locked=False,
    ):
        """

        :param where_conditions: "NOT (bala < %s)"
        :param where_values: (5, 10,)
        """
        with self._engine.session_manager(session=session) as s:
            if cache is True:
                return s.cache.query(
                    engine=self._engine,
                    table=self._table,
                    where_conditions=where_conditions,
                    where_values=where_values,
                    fallback=self._query,
                    limit=limit,
                    order_by=order_by,
                    locked=locked,
                )

            return self._query(
                where_conditions=where_conditions,
                where_values=where_values,
                session=s,
                limit=limit,
                order_by=order_by,
                locked=locked,
            )

    @base.error_catcher
    def count(self, session=None, filters=None):
        with self._engine.session_manager(session=session) as s:
            result = self._table.count(engine=self._engine, session=s, filters=filters)
            data = list(result.fetchall())
            return data[0]["count"]


class UndefinedAttribute(common_exc.RestAlchemyException):
    message = "Class attribute %(attr_name)s must be provided."


class SQLStorableMixin(base.AbstractStorableMixin, metaclass=abc.ABCMeta):
    _saved = False

    _ObjectCollection = ObjectCollection

    __tablename__ = None

    @classmethod
    def get_table(cls):
        try:
            table = cls.__operational_storage__.get(
                tables.OPERATIONAL_STORAGE_SIMPLE_TABLE_KEY,
            )
        except common_exc.NotFoundOperationalStorageError:
            if cls.__tablename__ is None:
                raise UndefinedAttribute(attr_name="__tablename__")
            table = tables.SQLTable(
                engine=cls._get_engine(),
                table_name=cls.__tablename__,
                model=cls,
            )
            cls.__operational_storage__.store(
                tables.OPERATIONAL_STORAGE_SIMPLE_TABLE_KEY,
                table,
            )
        return table

    @classmethod
    def _get_engine(cls):
        return engines.engine_factory.get_engine()

    @classmethod
    def restore_from_storage(cls, **kwargs):
        model_format = {}
        for name, value in kwargs.items():
            model_format[name] = (
                cls.properties.properties[name]
                .get_property_type()
                .from_simple_type(value)
            )
        obj = cls.restore(**model_format)
        obj._saved = True
        return obj

    @base.error_catcher
    @base.dead_lock_catcher
    def insert(self, session=None):
        # TODO(efrolov): Add filters parameters.
        with self._get_engine().session_manager(session=session) as s:
            try:
                self.get_table().insert(
                    engine=self._get_engine(),
                    data=self._get_prepared_data(),
                    session=s,
                )
                # TODO(efrolov): Check result
            except exc.Conflict as e:
                raise exceptions.ConflictRecords(model=self, msg=str(e))
            self._saved = True

    def save(self, session=None):
        # TODO(efrolov): Add filters parameters.
        self.update(session) if self._saved else self.insert(session)

    @base.error_catcher
    @base.dead_lock_catcher
    def update(self, session=None, force=False):
        # TODO(efrolov): Add filters parameters.
        if self.is_dirty() or force:
            self.validate()
            with self._get_engine().session_manager(session=session) as s:
                try:
                    result = self.get_table().update(
                        engine=self._get_engine(),
                        ids=self._get_prepared_data(self.get_id_properties()),
                        data=self._get_prepared_data(self.get_data_properties()),
                        session=s,
                    )
                except exc.Conflict as e:
                    raise exceptions.ConflictRecords(model=self, msg=str(e))
                if result.get_count() == 0:
                    _filters = {
                        name: dm_filters.EQ(prop.value)
                        for name, prop in self.get_id_properties().items()
                    }
                    type(self).objects.get_one(filters=_filters, session=s)
                if result.get_count() > 1:
                    raise exceptions.MultipleUpdatesDetected(model=self, filters={})

    @base.error_catcher
    @base.dead_lock_catcher
    def delete(self, session=None):
        # TODO(efrolov): Add filters parameters.
        with self._get_engine().session_manager(session=session) as s:
            result = self.get_table().delete(
                engine=self._get_engine(),
                ids=self._get_prepared_data(self.get_id_properties()),
                session=s,
            )
            # TODO(efrolov): Check result
            return result

    @classmethod
    def to_simple_type(cls, value):
        if value is None:
            return None
        id_type = cls.get_id_property().popitem()[-1].get_property_type()
        if isinstance(value, models.Model):
            return id_type.to_simple_type(value.get_id())
        # Allow to filter by id without full model
        return id_type.to_simple_type(value)

    @classmethod
    @base.dead_lock_catcher
    def from_simple_type(cls, value):
        if value is None:
            return None
        if isinstance(value, base.PrefetchResult):
            for name in cls.id_properties.keys():
                if value[name]:
                    break
            else:
                return None
            return cls.restore_from_storage(**value)
        for name in cls.id_properties:
            value = (
                cls.properties.properties[name]
                .get_property_type()
                .from_simple_type(value)
            )
            engine = engines.engine_factory.get_engine()
            return cls.objects.get_one(
                filters={name: dm_filters.EQ(value)}, cache=engine.query_cache
            )


class SQLStorableWithJSONFieldsMixin(SQLStorableMixin, metaclass=abc.ABCMeta):
    """Use only if database's client doesn't support JSON fields natively."""

    __jsonfields__ = None

    @classmethod
    def restore_from_storage(cls, **kwargs):
        if cls.__jsonfields__ is None:
            raise UndefinedAttribute(attr_name="__jsonfields__")
        kwargs = kwargs.copy()
        for field in cls.__jsonfields__:
            # Some databases' clients support JSON fields natively.
            if isinstance(kwargs[field], str):
                kwargs[field] = orjson.loads(kwargs[field])
        return super(SQLStorableWithJSONFieldsMixin, cls).restore_from_storage(**kwargs)

    def _get_prepared_data(self, properties=None):
        if self.__jsonfields__ is None:
            raise UndefinedAttribute(attr_name="__jsonfields__")
        result = super(SQLStorableWithJSONFieldsMixin, self)._get_prepared_data(
            properties
        )
        if properties is None:
            json_properties = self.__jsonfields__
        else:
            json_properties = set(self.__jsonfields__).intersection(
                set(properties.keys())
            )
        for field in json_properties:
            result[field] = orjson.dumps(
                result[field], option=orjson.OPT_NON_STR_KEYS
            ).decode()
        return result



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

_DATABASE_URI_DEFAULT = "mysql://test:test@127.0.0.1:/test"

# Use get_database_uri() if you want xdist workers to parallelize!
DATABASE_URI = os.getenv("DATABASE_URI", _DATABASE_URI_DEFAULT)


def get_database_uri():
    return os.getenv("DATABASE_URI", _DATABASE_URI_DEFAULT)



# Copyright 2020 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.dm import models
from restalchemy.dm import relationships
from restalchemy.tests.unit import base


class MyModel(models.Model):
    pass


class RelationshipTestCase(base.BaseTestCase):
    def test_init_incorrect_value(self):
        with self.assertRaises(TypeError):
            relationships.Relationship(property_type=MyModel, value="IncorrectValue")

    def test_init_with_correct_value(self):
        self.assertIsInstance(
            relationships.Relationship(property_type=MyModel, value=MyModel()),
            relationships.Relationship,
        )



# Copyright 2021 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import

import logging

from restalchemy.api import middlewares

LOG = logging.getLogger(__name__)


# USE IT BEFORE CONTEXT MIDDLEWARE!
class RetryOnErrorsMiddleware(middlewares.Middleware):
    def __init__(self, application, exceptions, max_retry=3):
        self._exceptions = exceptions
        self._max_retry = max_retry if max_retry > 0 else 1
        super(RetryOnErrorsMiddleware, self).__init__(application)

    def process_request(self, req):
        retry_count = 0
        while True:
            retry_count += 1
            try:
                return req.get_response(self.application)
            except self._exceptions as e:
                LOG.warning(
                    "Unable to process request: %d/%d %s",
                    retry_count,
                    self._max_retry,
                    e,
                )
                if retry_count >= self._max_retry:
                    raise



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstractMigrationStep):
    def __init__(self):
        self._depends = [""]

    @property
    def migration_id(self):
        return "0d06a988-90cc-48ab-a842-b979cdf8975d"

    def upgrade(self, session):
        pass

    def downgrade(self, session):
        pass


migration_step = MigrationStep()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import requests

# -----------------------------------------------------------------------------
# List API versions
# -----------------------------------------------------------------------------

url = "http://127.0.0.1:8000/"
headers = {"cache-control": "no-cache"}
print("Make GET request to %s" % url)
response = requests.request("GET", url, headers=headers)
print("Response is %s. Headers is %s" % (response.text, response.headers))


# -----------------------------------------------------------------------------
# Create Foo resource
# -----------------------------------------------------------------------------

url = "http://127.0.0.1:8000/foos/"
payload = {"foo-field1": 999, "foo-field2": "foo obj"}
headers = {"content-type": "application/json", "cache-control": "no-cache"}
print("Make POST request to %s with payload %s" % (url, payload))
response = requests.request("POST", url, json=payload, headers=headers)
print("Response is %s. Headers is %s" % (response.text, response.headers))

foo_uuid = response.json()["uuid"]


# -----------------------------------------------------------------------------
# Get list of Foo resources
# -----------------------------------------------------------------------------

url = "http://127.0.0.1:8000/foos/"
headers = {"cache-control": "no-cache"}
print("Make GET (list on collection) request to %s" % url)
response = requests.request("GET", url, headers=headers)
print("Response is %s. Headers is %s" % (response.text, response.headers))


# -----------------------------------------------------------------------------
# Get Foo resource by uuid
# -----------------------------------------------------------------------------

url = "http://127.0.0.1:8000/foos/%s" % foo_uuid
headers = {"cache-control": "no-cache"}
print("Make GET request to foo resource %s" % url)
response = requests.request("GET", url, headers=headers)

print("Response is %s. Headers is %s" % (response.text, response.headers))


# -----------------------------------------------------------------------------
# Create Bar resource
# -----------------------------------------------------------------------------

url = "http://127.0.0.1:8000/foos/%s/bars/" % foo_uuid

payload = {"bar-field1": "test bar"}
headers = {
    "content-type": "application/json",
    "cache-control": "no-cache",
}
print("Make POST request to %s with payload %s" % (url, payload))
response = requests.request("POST", url, json=payload, headers=headers)
print("Response is %s. Headers is %s" % (response.text, response.headers))


bar_uuid = response.json()["uuid"]


# -----------------------------------------------------------------------------
# Delete Bar resource
# -----------------------------------------------------------------------------

url = "http://127.0.0.1:8000/bars/%s" % bar_uuid
headers = {"cache-control": "no-cache"}
print("Make DELETE request to %s" % url)
response = requests.request("DELETE", url, headers=headers)
print("Done!")
